# Analysis results for {{ filename }}
{% if issues %}
{% for issue in issues %}

## {{ issue.title }}
- SWC ID: {{ issue.swc_id }}
- Type: {{ issue.type }}
- Contract: {{ issue.contract | default("Unknown") }}
- Function name: `{{ issue.function }}`
- PC address: {{ issue.address }}

### Description

{{ issue.description.rstrip() }}
{% if issue.filename and issue.lineno %}
In file: {{ issue.filename }}:{{ issue.lineno }}
{% endif %}
{% if issue.code %}

### Code

```
{{ issue.code }}
```
{% endif %}
{% if verbose and issue.debug %}
--------------------
### Debugging Information:

{{ issue.debug }}

{% endif %}
{% endfor %}
{% else %}

The analysis was completed successfully. No issues were detected.
{% endif %}