# We use RsT document formatting in docstring. For example :param to mark parameters.
# See PEP 287
__docformat__ = 'restructuredtext'

# Accept mythril.VERSION to get mythril's current version number
from .version import VERSION  # NOQA
